
#include <lubishev/Exception.h>
#include <lubishev/Scene.h>

#include <ctime>
#include <iostream>
#include <QApplication>
#include <QMessageBox>

namespace lubishev {

Scene::Scene(DB *db, QWidget *parent) : QWidget(parent),
                                        mEventStart(0),
                                        mEventEnd(0),
                                        mEventID(0),
                                        mLastHoveredEventID(0),
                                        mCursorTime(0),
                                        mSnap(300),
                                        mZoom(false),
                                        mZoomInterval(3600),
                                        mClickStart(0),
                                        mClick(false),
                                        mDB(db) {
    mPen.setBrush(QBrush(Qt::SolidPattern));
    setMouseTracking(true);
    setFocusPolicy(Qt::StrongFocus);
    mEdit = new Edit(this);
    mCalendar = new QCalendarWidget(this);
    mCalendar->setVisible(false);
    connect(mCalendar, SIGNAL(clicked(const QDate&)),
            this,      SLOT(onCalendarClick(const QDate&)));
}

Scene::~Scene() {
    delete mCalendar;
    delete mEdit;
    clearEvents();
}

bool Scene::calendarIsVisible() const {
    return mCalendar->isVisible();
}

void Scene::clearEvents() {
    for (ESEIt it = mESE.begin(); it != mESE.end(); it++) {
        delete it->second;
        delete it->first;
    }
    mESE.clear();
}

float Scene::secondHeight() const {
    return mSecondHeight;
}

void Scene::setCalendarVisible(bool state) {
    mCalendar->setVisible(state);
}

void Scene::setInterval(time_t start, time_t end) {
    mStart = start;
    mEnd   = end;
    refreshInterval();
}

void Scene::setInterval(const QDate &date) {
    QDateTime qnow;
    qnow.setTime_t(mStart);
    time_t diff = qnow.daysTo(QDateTime(date)) * 86400;
    setInterval(mStart + diff, mEnd + diff);
}

void Scene::setIntervalNow() {
    time_t now = time(0);
    setInterval(now - mZoomInterval, now + mZoomInterval);
}

u16 Scene::snap() const {
    return mSnap;
}

float Scene::timeToY(time_t tm) {
    return (tm - mStart) * mSecondHeight;
}

void Scene::onCalendarClick(const QDate &date) {
    setInterval(date);
    setCalendarVisible(false);
}

void Scene::keyPressEvent(QKeyEvent *e) {
    if (mEdit->isVisible())
        return;
    bool doUpdate = false;
    switch (e->key()) {
        case Qt::Key_X:
            // Mouse hold + X.
            if (mClick && SceneEvent::selected()) {
                try {
                    SceneEvent::selected()->setSelectedPartTime(snapTime(time(0)));
                    SceneEvent::deselectSelected();
                }
                catch (Exception &e) {
                    QMessageBox::warning(this, e.getType(), e.getDesc());
                }
            }
            // Shift + X.
            else if (e->modifiers() && Qt::ShiftModifier) {
                time_t now = snapTime(time(0));
                createEvent(now, now + mSnap);
            }
            setIntervalNow();
            update();
            break;
        case Qt::Key_Tab:
            if (hoverEvent(true))
                doUpdate = true;
            break;
        case Qt::Key_Control:
            mZoom = true;
            break;
        case Qt::Key_Shift:
            mEventStart = 0;
            mEventEnd = 0;
            break;
        default:
            break;
    }
    if (doUpdate)
        update();
}

void Scene::keyReleaseEvent(QKeyEvent *e) {
    if (mEdit->isVisible())
        return;
    switch (e->key()) {
        case Qt::Key_Control:
            mZoom = false;
            break;
        case Qt::Key_Shift:
            mEventStart = 0;
            mEventEnd = 0;
            break;
        default:
            break;
    }
}

void Scene::mouseDoubleClickEvent(QMouseEvent *e) {
    if (mEdit->isVisible())
        return;
    SceneEvent *se = SceneEvent::selected();
    if (se)
        mEdit->display(se);
}

void Scene::mouseMoveEvent(QMouseEvent *e) {
    if (mEdit->isVisible())
        return;
    bool doUpdate = refreshTimeLine(e->pos());
    if (!mClick) {
        SceneEvents sceneEvents = xyToEvents(e->pos());
        if (!sceneEventsEqual(sceneEvents, mLastHovered))
            mLastHoveredEventID = 0;
        mLastHovered = sceneEvents;
        if (sceneEvents.size()) {
            if (hoverEvent())
                doUpdate = true;
        }
        else if (SceneEvent::mouseAbove()) {
            SceneEvent::dehoverMouseAbove();
            doUpdate = true;
        }
    }
    else {
        SceneEvent *se = SceneEvent::selected();
        if (se && mClickStart != mCursorTime) {
            time_t diff = mCursorTime - mClickStart;
            mClickStart = mCursorTime;
            doEventDiff(se, diff);
            // If the event was shortened over the limits, restore it.
            if (se->event()->end() - se->event()->start() < mSnap)
                doEventDiff(se, -diff);
            se->refreshRect();
            doUpdate = true;
        }
    }
    if (doUpdate)
        update();
}

void Scene::mousePressEvent(QMouseEvent *e) {
    if (mEdit->isVisible())
        return;
    if (e->button() == Qt::LeftButton) {
        if (e->modifiers() && Qt::ShiftModifier)
            mEventStart = mCursorTime;
        else {
            tryToSelectEvent();
            if (SceneEvent::selected()) {
                mClick = true;
                mClickStart = mCursorTime;
            }
        }
    }
}

void Scene::mouseReleaseEvent(QMouseEvent *e) {
    if (mEdit->isVisible())
        return;
    bool doUpdate = false;
    if (e->button() == Qt::LeftButton) {
        if (e->modifiers() && Qt::ShiftModifier && mEventStart) {
            mEventEnd = mCursorTime;
            //mEventEnd = yToTime(e->pos().y());
            // Make sure start time < end time.
            if (mEventStart > mEventEnd) {
                time_t tmp = mEventStart;
                mEventStart = mEventEnd;
                mEventEnd   = tmp;
            }
            // Create new event only if it's longer than 5 minutes.
            if (mEventEnd - mEventStart >= mSnap) {
                createEvent(mEventStart, mEventEnd);
                //Event *e = new Event(mDB, mEventStart, mEventEnd);
                //mESE[e] = 0;
                doUpdate = true;
            }
            mEventStart = 0;
            mEventEnd   = 0;
        }
        mClick = false;
    }
    if (doUpdate) {
        refreshInterval();
        update();
    }
}

void Scene::paintEvent(QPaintEvent *e) {
    QPainter p(this);
    p.setPen(mPen);
    setColor(p, QColor(255, 0, 0));
    p.fillRect(rect(), QColor(255, 255, 255));
    for (VHIt it = mVisibleHours.begin(); it != mVisibleHours.end(); it++) {
        time_t h = *it;
        float y = timeToY(h);
        setColor(p, QColor(255, 0, 0, 50));
        p.drawLine(QPointF(0, y), QPointF(width(), y));
        QRectF hourRect = QRectF(QPointF(0, y),
                                 QPointF(width(), y + mSecondHeight * 3600));
        struct tm * ti = localtime(&h);
        // Hour.
        setFontSize(p, mSecondHeight * 3600);
        p.drawText(hourRect,
                   Qt::AlignHCenter | Qt::AlignVCenter,
                   QString("%1").arg(ti->tm_hour));
    }
    // Set text colour.
    /*
    QPalette pal = QApplication::palette();
    QColor col = pal.color(QPalette::Base);
    pal.setColor(QPalette::Text, QColor(255, 0, 0));
    QApplication::setPalette(pal);//, "QTextDocument");
    */
    for (ESEIt it = mESE.begin(); it != mESE.end(); it++)
        if (it->second && it->second != SceneEvent::mouseAbove())
            it->second->draw(&p);
    // Draw hovered event the last.
    if (SceneEvent::mouseAbove())
        SceneEvent::mouseAbove()->draw(&p);
    // Restore text colour.
    //pal.setColor(QPalette::Text, col);
    //QApplication::setPalette(pal, "QTextDocument");
    // Time line.
    if (mCursorTime) {
        setColor(p, QColor(0, 0, 0, 50));
        float y = timeToY(mCursorTime);
        p.drawLine(QPointF(0, y), QPointF(width(), y));
        setFontSize(p, mSecondHeight * 380);
        p.drawText(QPointF(width() / 2, y), timeToString(mCursorTime));
    }
}

void Scene::resizeEvent(QResizeEvent *e) {
    refreshInterval();
    QWidget::resizeEvent(e);
    mEdit->resize(e->size());
    mCalendar->resize(e->size());
    QFont font = mEdit->document()->defaultFont();
    font.setPointSize(secondHeight() * 190);
    mEdit->document()->setDefaultFont(font);
}

void Scene::wheelEvent(QWheelEvent *e) {
    if (mEdit->isVisible())
        return;
    // TODO: settings: multiple of N minutes to scroll.
    int delta = e->delta() * 8;
    if (mZoom) {
        mZoomInterval -= delta;
        if (mZoomInterval < 1800)
            mZoomInterval = 1800;
        else if (mZoomInterval > 18000)
            mZoomInterval = 18000;
        setIntervalNow();
    }
    else {
        mStart -= delta;
        mEnd -= delta;
        refreshInterval();
    }
    update();
}

void Scene::createEvent(time_t start, time_t end) {
    Event *e = new Event(mDB, start, end);
    mESE[e] = 0;
}

void Scene::doEventDiff(SceneEvent *se, time_t diff) {
    Event *e = se->event();
    switch (se->containsResult()) {
        case SceneEvent::EVENT:
            e->setTimes(e->start() + diff, e->end() + diff);
            break;
        case SceneEvent::HANDLE_TOP:
            e->setTimes(e->start() + diff, e->end());
            break;
        case SceneEvent::HANDLE_BOTTOM:
            e->setTimes(e->start(), e->end() + diff);
            break;
        default:
            // Can't be, because it's been selected.
            // Inserted to just to suppress compiler warning.
            break;
    }
}

bool Scene::hoverEvent(bool next) {
    if (next)
        mLastHoveredEventID++;
    if (mLastHoveredEventID >= mLastHovered.size())
        mLastHoveredEventID = 0;
    if (mLastHovered.size()) {
        mLastHovered[mLastHoveredEventID]->setMouseAbove(true);
        return true;
    }
    return false;
}

time_t Scene::nextHour(time_t start, time_t end) {
    u16 distance = start % 3600;
    if (!distance)
        return start;
    else if (start + 3600 - distance <= end)
        return start + 3600 - distance;
    else
        return 0;
}

void Scene::refreshCell() {
    mVisibleHours.clear();
    time_t h = mStart;
    bool firstRun = true;
    while ((h = nextHour(h, mEnd))) {
        if (firstRun) {
            mVisibleHours.push_back(h - 3600);
            firstRun = false;
        }
        mVisibleHours.push_back(h);
        // Go one second after found hour.
        h++;
    }
}

void Scene::refreshEvents() {
    QStringList ids = mDB->eventsWithin(mStart, mEnd);
    // Find out events outside interval.
    Events toRemove;
    for (ESEIt it = mESE.begin(); it != mESE.end(); it++) {
        bool found = false;
        for (QStringList::iterator jt = ids.begin(); jt != ids.end(); jt++)
            if (it->first->id() == *jt) {
                found = true;
                ids.erase(jt);
                break;
            }
        if (!found)
            toRemove.push_back(it->first);
    }
    // Remove them.
    for (EIt it = toRemove.begin(); it != toRemove.end(); it++) {
        SceneEvent *se = mESE[*it];
        mESE.erase(mESE.find(*it));
        if (se)
            delete se;
        delete *it;
    }
    // Add new events. Not graphical ones yet.
    for (QStringList::iterator it = ids.begin(); it != ids.end(); it++) {
        Event *e = new Event(mDB, *it);
        mESE[e] = 0;
    }
    // Create missing graphic ones and refresh new ones.
    for (ESEIt it = mESE.begin(); it != mESE.end(); it++)
        if (it->second)
            it->second->refreshRect();
        else {
            SceneEvent *se = new SceneEvent(this, it->first);
            it->second = se;
        }
}

void Scene::refreshInterval() {
    mSecondHeight = static_cast<float>(height()) / (mEnd - mStart);
    // The scene has not yet been drawn.
    if (!mSecondHeight)
        return;
    refreshCell();
    refreshEvents();
    // Update Calendar.
    QDateTime dt;
    dt.setTime_t(mStart);
    mCalendar->blockSignals(true);
    mCalendar->setSelectedDate(dt.date());
    mCalendar->blockSignals(false);
}

bool Scene::refreshTimeLine(const QPointF &pos) {
    time_t lineNow = snapTime(yToTime(pos.y()));
    if (lineNow != mCursorTime) {
        mCursorTime = lineNow;
        emit cursorTimeChanged(mCursorTime);
        return true;
    }
    return false;
}

bool Scene::sceneEventsEqual(const SceneEvents &se1, const SceneEvents &se2) {
    if (se1.size() != se2.size())
        return false;
    SECIt it = se1.begin();
    SECIt jt = se2.begin();
    for (; it != se1.end(); it++, jt++)
        if (*it != *jt)
            return false;
    return true;
}

time_t Scene::snapTime(time_t t) {
    u16 distance = t % mSnap;
    if (!distance)
        return t;
    else 
        return t + mSnap - distance;
}

void Scene::tryToSelectEvent() {
    bool doUpdate = false;
    // Select the hovered one.
    if (SceneEvent::mouseAbove()) {
        SceneEvent::mouseAbove()->setSelected(true);
        doUpdate = true;
    }
    else if (SceneEvent::selected()) {
        SceneEvent::deselectSelected();
        doUpdate = true;
    }
    if (doUpdate)
        update();
}

Scene::SceneEvents Scene::xyToEvents(const QPointF &pos) {
    SceneEvents sceneEvents;
    // The hovered SceneEvent always gets first into the list.
    SceneEvent *above = SceneEvent::mouseAbove();
    if (above && above->contains(pos))
        sceneEvents.push_back(above);
    for (ESEIt it = mESE.begin(); it != mESE.end(); it++)
        if (it->second && it->second != above && it->second->contains(pos))
            sceneEvents.push_back(it->second);
    return sceneEvents;
}

time_t Scene::yToTime(float y) {
    return mStart + y / mSecondHeight;
}

} // namespace lubishev

